;PIC18F2450USBvb
;mei2008
;Generiek USB-HID firmware full speed 64 bytes, voor PIC18F2450/4450 en 4550
;kristal 20 MHz
;met voorbeeld-routines om poorten te lezen en te schrijven met de VB6-toepassing PICbesturingUSB
;VID=&HBC0B, PID=&H1944
; 
; THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION.  NO WARRANTIES, WHETHER 
; EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED 
; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY 
; TO THIS SOFTWARE. THE AUTHOR SHALL NOT, UNDER ANY CIRCUMSTANCES, BE LIABLE 
; FOR SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
; 	De USB-routines zijn nagenoeg geheel gebaseerd op voorbeelden van Jan Axelson (GENHID) en 
;	Brad Minch (lab2) en bijdragen van het PIC- en CO-forum. Het programma of delen ervan mag/mogen niet worden verhandeld 
;	en evenmin worden toegepast in commerciele producten. De VID en PID zijn niet voorzien
;	van rechten en mogen daarom alleen worden gebruikt voor particuliere (hobby)-doeleinden.








#include <p18F2450.inc>
#include "usb_defs.inc"
#include "ENGR2210.inc"

#define	bInterval			0x01
#define	NUM_CONFIGURATIONS	1
#define	NUM_INTERFACES		1
#define 	NUM_STRINGS			2
#define 	EP0_packet_size		.8	; maximum packet size for EP0 is 8 bytes,
#define 	MAX_packet_size		.64	; for full-speed peripherals it can be 8, 16, 32, or 64 bytes
#define 	USB_Buffer			0x460		; Register location after some buffer descriptor registers
; PIC18F2450/4450 eerste 96 bytes=ACCESSram, 0x100 start 1 bank RAM, 0x400 start 1 bank USBram
; PIC18F4550 eerste 96 bytes=ACCESSram, 0x100 start 3 banken RAM, 0x400 start 4 banken USBram

;de als commentaar gekenmerkte configuraties gelden voor een PIC18F4550 
	config	FOSC = HSPLL_HS      ;HS oscillator, PLL enabled, HS used by USB
	config 	PLLDIV = 5				;20 MHz klok
	config 	CPUDIV = OSC1_PLL2	;./. 2, fosc = 48 MHz 
	config 	USBDIV = 2			;PLL/2, USBklok = 48 MHz; full speed
	config 	FCMEN = OFF
	config 	IESO = OFF
	config 	PWRT = OFF
	config 	BOR = OFF
;	config BORV = 3
	config 	VREGEN = ON
	config 	WDT = OFF
	config 	WDTPS = 32768
	config 	MCLRE = ON
	config 	LPT1OSC = OFF
	config 	PBADEN = OFF
;	config CCP2MX = ON
	config 	STVREN = ON		;USB spanning
	config 	LVP = OFF
	config 	ICPRT = OFF
	config 	XINST = OFF
	config 	DEBUG = OFF
	config 	CP0 = OFF
	config 	CP1 = OFF
;	config CP2 = OFF
;	config CP3 = OFF
	config 	CPB = OFF
;	config CPD = OFF
	config 	WRT0 = OFF
	config 	WRT1 = OFF
;	config WRT2 = OFF
;	config WRT3 = OFF
	config 	WRTB = OFF
	config 	WRTC = OFF
;	config WRTD = OFF
	config 	EBTR0 = OFF
	config 	EBTR1 = OFF
;	config EBTR2 = OFF
;	config EBTR3 = OFF
	config 	EBTRB = OFF

;----------------------------------------------------------------------------
;Variables	
ACCESSdata	 	udata_acs 	0x000	;96 bytes accessRAM
	
Flags		res	1		;byte for indicator flag bits
Tempkar	res	1		;karakter
Tempnib	res	1
Tempwreg	res	1		;save W
Adrhig	res	1
Adrlow	res	1
COMMOD	res	1		;POORT-instructie save
POORTN	res	1		;POORT-instructie save
KANBYT	res	1		;POORT-instructie save
Teller	res	1
WREG_TEMP	res	1		;to save WREG during interrupt
STATUS_TEMP	res	1		;to save STATUS during interrupt

bank1			udata		0x100	;banked ram
USB_buffer_desc	res		4
USB_error_flags	res 		1
USB_curr_config	res		1
USB_device_status	res		1
USB_protocol	res		1
USB_idle_rate	res		1
USB_dev_req		res		1
USB_address_pending	res	1
USB_desc_ptr	res		1
USB_bytes_left	res		1
USB_loop_index	res		1
USB_packet_length	res		1
USB_USTAT		res		1
USB_USWSTAT		res		1
Report_type		res		1
Feature_buffer	res		2
Ctrl_io_buffer	res		2
IntIbuffer		res		MAX_packet_size
IntObuffer		res		MAX_packet_size
USB_buffer_data	res		EP0_packet_size


STARTUP		code		0x0000
		goto		Main					; Reset vector
		nop
		nop
		goto		$						; High-priority interrupt vector trap
		nop
		nop
		nop
		nop
		nop
		nop
		goto		$						; Low-priority interrupt vector trap

APPLICATION		code
Main
		call		Initiatie
MainLoop	call		SCAN			;kijk of er input is
		iorlw		0x00			;als W=0 niet, blijf kijken
		bz		MainLoop
	
		banksel	IntIbuffer
		lfsr		FSR1, IntIbuffer	;zet pointer op IntIbuffer
							;begin van ontvangen commando

		call	READ				;haal karakter
		movlw	"P"				;poortcommando
		subwf	Tempkar,W
		bz	POORT	
		;movlw "X"				check vlgd commando
		;subwf Tempkar,W
		;bz	Xroutine			
		bra	MainLoop


POORT		goto	POORTcom
;Xroutine

;-------------------------------------------------------------		
POORTcom	rcall	READ		;haal poortnaam
		movwf	POORTN
		rcall	READ		;commandotype
		movwf	COMMOD
						
EXEC		movlw	"I"		;input?
		subwf	COMMOD,W
		bz	SETIN
		movlw	"O"		;output?
		subwf	COMMOD,W
		bz	SETOUT
		movlw	"L"		;lezen?
		subwf	COMMOD,W
		bz	READKAN
		movlw	"S"		;schrijven?
		subwf	COMMOD,W
		bz	WRITKAN
		movlw	"Z"		;mix in/out?
		subwf	COMMOD,W
		bz	ZETKAN
		bra	MainLoop


ZETKAN	movlw	"A"		;zet betreffende bits als in- of output
		subwf	POORTN,W
		bz	ZETKANA
		movlw	"B"		;
		subwf	POORTN,W
		bz	ZETKANB
		bra	MainLoop
		
ZETKANA	rcall INBYT
		movwf	TRISA
		bra	MainLoop
ZETKANB	rcall INBYT
		movwf	TRISB
		bra	MainLoop
ZETKANC	rcall INBYT
		movwf	TRISC
		bra	MainLoop

SETIN		movlw	"A"		;zet betreffende poort als input
		subwf	POORTN,W
		bz	SETINA
		movlw	"B"		;
		subwf	POORTN,W
		bz	SETINB
		bra	MainLoop
		
SETINA	movlw	0xFF
		movwf	TRISA
		bra	MainLoop
SETINB	movlw	0xFF
		movwf	TRISB
		bra	MainLoop
		
SETOUT	movlw	"A"		;zet betreffende poort als output
		subwf	POORTN,W
		bz	SETOUTA
		movlw	"B"		;
		subwf	POORTN,W
		bz	SETOUTB
		bra	MainLoop
		
SETOUTA	clrf	TRISA
		bra	MainLoop
SETOUTB	clrf	TRISB
		bra	MainLoop
SETOUTC	clrf	TRISC
		bra	MainLoop
		
READKAN	lfsr	FSR2, IntObuffer
		movlw	"A"		;
		subwf	POORTN,W
		bz	READKANA
		movlw	"B"		;
		subwf	POORTN,W
		bz	READKANB
		bra	MainLoop
		
READKANA	movf	PORTA,W		;lees betreffende poort uit naar transmitter
		rcall	OUTHEX
		call	OUT8B
		bra	MainLoop
READKANB	movf	PORTB,W
		rcall	OUTHEX
		call	OUT8B
		bra	MainLoop
		
WRITKAN	movlw	"A"		;schrijf waarde in betreffende poort
		subwf	POORTN,W
		bz	WRITKANA
		movlw	"B"		;
		subwf	POORTN,W
		bz	WRITKANB
		bra	MainLoop
		
WRITKANA	rcall	INBYT
		movwf	LATA
		bra	MainLoop
WRITKANB	rcall	INBYT
		movwf	LATB
		bra	MainLoop
;----------------------------------------------------------------------------


;----------------------------------------------------------------------------
; -----------------  in- en uitvoerroutines  --------------------------------

OUTBYT		;output karakter in W, bewaart W in Tempkar
		movwf	Tempkar
		bra	OUTDIS

CRLOW			;output CR, bewaart W
		movwf	Tempwreg
		movlw	0Dh
		rcall	OUTDIS
		movf	Tempwreg,W
		return
		
OUTHEX		;output byte in W als 2 karakters
		movwf	Tempwreg
		movwf	Tempnib
		swapf	Tempwreg,W
		rcall	NOUT
		movf	Tempnib,W
NOUT		andlw	0x0F
		addlw	0x30
		movwf	Tempkar
		movlw	0x3A
		subwf	Tempkar,W
		bnc	OUTDIST
		movf	Tempkar,W
		addlw	0x07
		movwf	Tempkar
OUTDIST	movf	Tempkar,W		
OUTDIS	movwf	POSTINC2	;zet byte in IntObuffer
		return
;----------------------------------------------------------------------------			
INBYT			;lees 2 karakters zet als byte in W en Tempnib
		rcall	RDNIB
		movwf	Tempwreg
		swapf	Tempwreg,W
		andlw	0xF0		;high nibble
		movwf	Tempnib
		rcall	RDNIB
		addwf	Tempnib,f
		movf	Tempnib,W
		return	
		
RDNIB			;lees 1 karakter en zet als HEX nibble in W
		rcall	READ
		movwf	Tempwreg
		movlw	0x3A
		subwf	Tempwreg,W
		bnc	NOC
		movf	Tempwreg,W
		andlw	0Fh
		addlw	09h
		return
NOC		movf	Tempwreg,W
		andlw	0Fh
		return	
		
INADRES		;lees 2 bytes en zet in adresfiles
		rcall	INBYT
		movwf	Adrhig
		rcall	INBYT
		movwf	Adrlow
		return		

READ			;haalt karakter uit inputbuffer en hoogt pointer op
			;zet om in capitals, zet in W en Tempkar
		banksel IntIbuffer
		movf	POSTINC1, W
		andlw	7Fh
		movwf	Tempwreg
		movlw	61h		;a
		subwf	Tempwreg,W
		bnc	NOLO
		movf	Tempwreg,W
		andlw	5Fh
		movwf	Tempkar
		return
NOLO		movf	Tempwreg,W
		movwf	Tempkar
		return
		
READC			;haalt lower nibble van karakter, zet in W en Tempkar
		rcall	READ		;haal karakter in W
		andlw	0Fh		;lower nibble
		movwf	Tempkar
		return	


;----------------------------------------------------------------------------
;----------------------- Wachtloops --- worden nu niet gebruikt--------------------------
;	gebaseerd op fosc = 48 MHz, d.w.z. 12 MHz CPU-klok, 256 prescaler, eenheid=21,3 us
WALO1		
		;1 mS
		clrf	TMR0L
WALOw1	movlw	.47		;47 X 21,3 uS
		subwf	TMR0L,W
		bnz	WALOw1
		return
		
WALO5		
		;5 mS
		clrf	TMR0L
WALOw5	movlw	.235		;235 X 21,3 uS
		subwf	TMR0L,W
		bnz	WALOw5
		return

WALOU		;wacht 128 uS
		clrf	TMR0L
WALOwu	movlw	.6		;6 X 21,3 uS
		subwf	TMR0L,W
		bnz	WALOwu
		return



;======================================================================================================
Initiatie					
		call		InitUSB				; initialize the USB registers and serial interface engine
		repeat
			call		ServiceUSB			; service USB requests...
			banksel	USB_USWSTAT
		until USB_USWSTAT, ==, CONFIG_STATE		; ...until the host configures the peripheral
		banksel		Feature_buffer
		clrf		Feature_buffer, BANKED
		clrf		Feature_buffer+1, BANKED
		clrf		Ctrl_io_buffer, BANKED
		clrf		Ctrl_io_buffer+1, BANKED
		clrf		IntIbuffer, BANKED		;alleen eerste byte
		clrf		IntObuffer, BANKED
		call		SetupTimers				;set up timers
		call		SetupADC				;set up ADconversie
		call		SetupDiv				;set up diversen
		call		SetupIO				;set up poorten
		return
		
;------------------------------------------------------------------------
SetupTimers:
							;Fosc=48 MHz. Instr.cycle=12 MHz
		movlw		0xC7			;TIMERperiode,  8 bits timer, 
		movwf		T0CON, ACCESS	; set prescaler for Timer0 for 1:256 scaling, so 21,3 uS/bit
		movlw		0xB0
		movwf		T1CON		;off, 16bits, 8 prescale
		clrf		T2CON		;off
		return
;------------------------------------------------------------------------
SetupADC:
		clrf		ADCON0		;disabled
		movlw		0x0F
		movwf		ADCON1		;alle poorten digital
		return

;-------------------------------------------------------------------------		
SetupDiv:
		clrf		HLVDCON		;disabled highvoltagedetect
		return
;-------------------------------------------------------------------------
SetupIO:	bcf	INTCON2,RBPU	;weak pull up poort B
		movlw	0xFF
		movwf	TRISC		;inputs
		movlw	0x00
		movwf	TRISB		;outputs
		movlw	0x9F
		movwf	TRISA
		clrf	LATA		;output latch 0
		clrf	LATB
		clrf	LATC
		return

;==========================================================================================================			
SCAN		;wacht tot eventuele transmissie heeft plaatsgevonden, W=0: geen data, W<>0: n bytes ontvangen
		;de n bytes staan in de IntIbuffer
Wascan
		call		ServiceUSB				; service USB requests...
		movlw		.50					;50X21,3=1050 us
		subwf		TMR0L, W
		bnc		Wascan				; ...until Timer0 is 50
		bcf		INTCON, T0IF, ACCESS		; clear Timer0 interrupt flag
		andlw		0x00					;zet W op 0, daarmee eindigt sub als geen ontvangst
		banksel	BD1OST
		ifclr BD1OST, UOWN, BANKED			; check to see if the PIC owns the EP1 OUT buffer
			movf		BD1OAH, W, BANKED		; put EP1 OUT buffer pointer...
			movwf		FSR0H, ACCESS
			movf		BD1OAL, W, BANKED
			movwf		FSR0L, ACCESS		; ...into FSR0
			movlw		low(IntIbuffer)
			movwf		FSR1L, ACCESS
			movlw		high(IntIbuffer)
			movwf		FSR1H, ACCESS
			movlw		MAX_packet_size-1
			movwf		Teller
			movf		POSTINC0, W			; get the byte in the buffer and...
			movwf		POSTINC1			; ...put it in the interrupt I/O buffer
			decf		Teller,f
			bnz		$-6
			movf		INDF0, W			; get the last byte in the buffer and...
			movwf		INDF1				; ...put it the interrupt I/O buffer
			banksel		BD1OST
			movlw		MAX_packet_size
			movwf		BD1OBC			; reset the EP1 OUT byte count
			movf		BD1OST, W, BANKED		; get the EP1 OUT status register
			andlw		0x40				; extract the DATA01 bit
			xorlw		0x40				; toggle the DATA01 bit
			iorlw		0x88				; set UOWN and DTS bits, tevens W<>0
			movwf		BD1OST, BANKED		; turn the buffer back over to USB
		endi	
						
		return						

OUT8B		;zendt bytes die staan in de IntObuffer
			repeat
				call		ServiceUSB		; service USB requests...
				banksel		BD1IST
			untilclr BD1IST, UOWN, BANKED		; ...until the PIC owns the EP1 IN buffer
			movf		BD1IAH, W, BANKED		; put EP1 IN buffer pointer...
			movwf		FSR0H, ACCESS
			movf		BD1IAL, W, BANKED
			movwf		FSR0L, ACCESS			; ...into FSR0
			movlw		low(IntObuffer)
			movwf		FSR1L, ACCESS
			movlw		high(IntObuffer)
			movwf		FSR1H, ACCESS
			movlw		MAX_packet_size-1
			movwf		Teller
			movf		POSTINC1, W			; get the byte in interrupt I/O buffer and...
			movwf		POSTINC0			; ...put it in the  buffer
			decf		Teller,f
			bnz		$-6
			movf		INDF1, W			; get the last byte in the interrupt I/O buffer and...
			movwf		INDF0				; ...put it the  buffer
			banksel		BD1IBC
			movlw		MAX_packet_size
			movwf		BD1IBC, BANKED			; set EP1 IN buffer byte count to MAX_packet_size
			movf		BD1IST, W, BANKED			; get EP1 IN status register
			andlw		0x40					; extract the DATA01 bit
			xorlw		0x40					; toggle the DATA01 bit
			iorlw		0x88					; set UOWN and DTS bits
			movwf		BD1IST, BANKED			; send packet
			return
			
			
						
USBSTUFF	code


			
Descriptors		code_pack			;18FXXX directive
Descriptor_begin
Device
			db			Configuration1-Device, DEVICE,0x10, 0x01,0x00, 0x00	; bLength, bDescriptorType, bcdUSB (low byte), bcdUSB (high byte),  bDeviceClass, bDeviceSubClass
			db			0x00, EP0_packet_size,0x0B, 0xBC,0x44, 0x19		; bDeviceProtocl, bMaxPacketSize, idVendor (low byte), idVendor (high byte)   &HBC0B, idProduct (low byte), idProduct (high byte)  &H1944
			db			0x01, 0x00, 0x01, 0x02, 0x03, NUM_CONFIGURATIONS	; bcdDevice (low byte), bcdDevice (high byte),iManufacturer, iProduct,iSerialNumber , bNumConfigurations
Configuration1
			db			Interface1-Configuration1, CONFIGURATION,0x29, 0x00,NUM_INTERFACES	; bLength, bDescriptorType,wTotalLength (low byte), wTotalLength (high byte),bNumInterfaces
			db			0x01, 0x00, 0xA0, 0x32						;bConfigurationValue, iConfiguration (none), bmAttributes, bMaxPower (100 mA)
Interface1			
			db			HID1-Interface1, INTERFACE, 0x00,0x00, 0x02		; bLength ,bDescriptorType, bInterfaceNumber,bAlternateSetting, bNumEndpoints (excluding EP0)
			db			0x03, 0x00, 0x00, 0x00						; bInterfaceClass (HID code), bInterfaceSubClass, bInterfaceProtocol, iInterface (none)
HID1
			db			Endpoint1-HID1, HID, 0x01, 0x01, 0x00			; bLength, bDescriptorType, bcdHID (low byte), bcdHID (high byte), bCountryCode (none)
			db			0x01, REPORT, Endreport1-Report1, 0x00				; bNumDescriptors, bDescriptorType, wDescriptorLength (low byte) lengte Report1, wDescriptorLength (high byte)
Endpoint1			
			db			Endpoint2-Endpoint1, ENDPOINT, 0x81, 0x03			; bLength, bDescriptorType, bEndpointAddress (EP1 IN), bmAttributes (Interrupt)
			db			MAX_packet_size, 0x00, bInterval				; wMaxPacketSize (low byte), wMaxPacketSize (high byte), bInterval 
Endpoint2
			db			Endendpoint2-Endpoint2, ENDPOINT, 0x01, 0x03			; bLength, bDescriptorType, bEndpointAddress (EP1 OUT), bmAttributes (Interrupt)
			db			MAX_packet_size, 0x00, bInterval				; wMaxPacketSize (low byte), wMaxPacketSize (high byte), bInterval
Endendpoint2			
			db			0x00								;vulling tot word
Report1
			db		  	0x06, 0xA0, 0xFF, 0x09, 0x01, 0xA1, 0x01		; Usage Page (Vendor Defined FFA0),Usage ID (Vendor Defined), Collection (Application),  Usage ID (Vendor Defined)
			db			0x09, 0x03							; The Input Report of  Usage ID (Vendor Defined)
			db			0x15, 0x00, 0x26, 0x00, 0xFF		    		; Logical Minimum (0), Logical Maximum (255)
			db			0x75, 0x08, 0x95, MAX_packet_size, 0x81, 0x02	; Report Size (8 bits),Report Count (nr fields), Input (Data, Variable, Absolute)
			db			0x09, 0x04							; The Output Report of  Usage ID (Vendor Defined) 
			db			0x15, 0x00, 0x26, 0x00, 0xFF				; Logical Minimum (0), Logical Maximum (255)
			db			0x75, 0x08, 0x95, MAX_packet_size, 0x91, 0x02 	; Report Size (8 bits), Report Count (nr fields), Output (Data, Variable, Absolute)
			db			0x09, 0x05							; The Feature Report of Usage ID (Vendor Defined) 
			db			0x15, 0x00, 0x26,	0x00, 0xFF				; Logical Minimum (0), Logical Maximum (255)
			db			0x75, 0x08, 0x95, 0x02, 0xB1, 0x02			; Report Size (8 bits), Report Count (2 fields), Feature (Data, Variable, Absolute),
			db			0xC0								; End Collection
Endreport1							
			db			0x00								;vulling tot word
String0
			db		String1-String0, STRING	; bLength, bDescriptorType
			db		0x09, 0x04			; wLANGID[0] (low byte), wLANGID[0] (high byte)
String1
			db		String2-String1, STRING	; bLength, bDescriptorType
			dw		'B','E','A','C','O','N'

String2
			db		String3-String2, STRING	; bLength, bDescriptorType
			dw		'U','S','B',' ','H','I','D',' ','I','/','O'

String3
			db      	String255-String3, STRING	 ; bLength, bDescriptorType
			dw      	'S','e','r','i','a','l','n','o',':','1','.','0'
String255
			db      	Descriptor_end-String255, STRING	 ; bLength, bDescriptorType
			dw      	'-'
Descriptor_end


InitUSB
			clrf		UIE, ACCESS				; mask all USB interrupts
			clrf		UIR, ACCESS				; clear all USB interrupt flags
			movlw		0x14
			movwf		UCFG, ACCESS			; configure USB for full-speed transfers and to use the on-chip transciever and pull-up resistor
			movlw		0x08
			movwf		UCON, ACCESS			; enable the USB module and its supporting circuitry
			banksel		USB_curr_config
			clrf		USB_curr_config, BANKED
			clrf		USB_idle_rate, BANKED
			clrf		USB_USWSTAT, BANKED		; default to powered state
			movlw		0x01
			movwf		USB_device_status, BANKED
			movwf		USB_protocol, BANKED		; default protocol to report protocol initially
			movlw		NO_REQUEST
			movwf		USB_dev_req, BANKED		; No device requests in process
			repeat								; do nothing...
			untilclr UCON, SE0, ACCESS			; ...until initial SE0 condition clears 
			return
;----------------------------------------------------------------------------------------------------
ServiceUSB		;controleert alle interruptbronnen van de USB
			select
				caseset	UIR, UERRIF, ACCESS
					clrf		UEIR, ACCESS		;clear all bits, resets UERRIF
					break
				caseset UIR, SOFIF, ACCESS
					bcf		UIR, SOFIF, ACCESS
					break
				caseset	UIR, IDLEIF, ACCESS
					bcf		UIR, IDLEIF, ACCESS
					bsf		UCON, SUSPND, ACCESS
					break
				caseset UIR, ACTVIF, ACCESS
					BCF UCON, SUSPND, ACCESS
					BTFSS UIR, ACTVIF, ACCESS
					BRA $+6
					BCF UIR, ACTVIF, ACCESS
					BRA $-6

					break
				caseset	UIR, STALLIF, ACCESS
					bcf		UIR, STALLIF, ACCESS
					break
				caseset	UIR, URSTIF, ACCESS		;RESET
					banksel		USB_curr_config
					clrf		USB_curr_config, BANKED
					bcf		UIR, TRNIF, ACCESS		; clear TRNIF four times to clear out the USTAT FIFO
					bcf 		UIR, TRNIF, ACCESS
					bcf		UIR, TRNIF, ACCESS
					bcf		UIR, TRNIF, ACCESS
					clrf		UEP0, ACCESS			; clear all EP control registers to disable all endpoints
					clrf		UEP1, ACCESS
					clrf		UEP2, ACCESS
					clrf		UEP3, ACCESS
					clrf		UEP4, ACCESS
					clrf		UEP5, ACCESS
					clrf		UEP6, ACCESS
					clrf		UEP7, ACCESS
					clrf		UEP8, ACCESS
					clrf		UEP9, ACCESS
					clrf		UEP10, ACCESS
					clrf		UEP11, ACCESS
					clrf		UEP12, ACCESS
					clrf		UEP13, ACCESS
					clrf		UEP14, ACCESS
					clrf		UEP15, ACCESS
					banksel		BD0OBC
					movlw		EP0_packet_size
					movwf		BD0OBC, BANKED
					movlw		low USB_Buffer			; EP0 OUT gets a buffer...
					movwf		BD0OAL, BANKED
					movlw		high USB_Buffer
					movwf		BD0OAH, BANKED			; ...set up its address
					movlw		0x88					; set UOWN bit (USB can write)
					movwf		BD0OST, BANKED
					movlw		low (USB_Buffer+EP0_packet_size)	; EP0 IN gets a buffer...
					movwf		BD0IAL, BANKED
					movlw		high (USB_Buffer+EP0_packet_size)
					movwf		BD0IAH, BANKED			; ...set up its address
					movlw		0x08					; clear UOWN bit (MCU can write)
					movwf		BD0IST, BANKED
					clrf		UADDR, ACCESS			; set USB Address to 0
					clrf		UIR, ACCESS				; clear all the USB interrupt flags
					movlw		ENDPT_CONTROL
					movwf		UEP0, ACCESS			; EP0 is a control pipe and requires an ACK
					movlw		0xFF					; enable all error interrupts
					movwf		UEIE, ACCESS
					banksel		USB_USWSTAT
					movlw		DEFAULT_STATE
					movwf		USB_USWSTAT, BANKED
					movlw		0x00
					movwf		USB_device_status, BANKED	; not self powered, remote wakeup disabled

					break
				caseset	UIR, TRNIF, ACCESS		;transactie klaar, kijk naar status	
					movlw		0x04
					movwf		FSR0H, ACCESS
					movf		USTAT, W, ACCESS
					andlw		0x7C					; mask out bits 0, 1, and 7 of USTAT
					movwf		FSR0L, ACCESS
					banksel		USB_buffer_desc
					movf		POSTINC0, W
					movwf		USB_buffer_desc, BANKED
					movf		POSTINC0, W
					movwf		USB_buffer_desc+1, BANKED
					movf		POSTINC0, W
					movwf		USB_buffer_desc+2, BANKED
					movf		POSTINC0, W
					movwf		USB_buffer_desc+3, BANKED
					movf		USTAT, W, ACCESS
					movwf		USB_USTAT, BANKED			; save the USB status register
					bcf		UIR, TRNIF, ACCESS		; clear TRNIF interrupt flag
					clrf		USB_error_flags, BANKED		; clear USB error flags
					movf		USB_buffer_desc, W, BANKED
					andlw		0x3C					; extract PID bits
					select
						case TOKEN_SETUP
							call		ProcessSetupToken
							break
						case TOKEN_IN
							call		ProcessInToken
							break
						case TOKEN_OUT
							call		ProcessOutToken
							break
					ends
					banksel USB_error_flags
					ifset USB_error_flags, 0, BANKED			; if there was a Request Error...
						banksel		BD0OBC
						movlw		EP0_packet_size
						movwf		BD0OBC				; ...get ready to receive the next Setup token...
						movlw		0x84
						movwf		BD0IST
						movwf		BD0OST				; ...and issue a protocol stall on EP0
					endi
					break
			ends
			return
;-------------------------------------------------------------------------------------------------------------------
ProcessSetupToken	
			banksel		USB_buffer_data
			movf		USB_buffer_desc+ADDRESSH, W, BANKED
			movwf		FSR0H, ACCESS
			movf		USB_buffer_desc+ADDRESSL, W, BANKED
			movwf		FSR0L, ACCESS
			movf		POSTINC0, W
			movwf		USB_buffer_data, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+1, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+2, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+3, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+4, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+5, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+6, BANKED
			movf		POSTINC0, W
			movwf		USB_buffer_data+7, BANKED
			banksel		BD0OBC
			movlw		EP0_packet_size
			movwf		BD0OBC, BANKED					; reset the byte count
			movlw		0x08
			movwf		BD0IST, BANKED					; return the in buffer to us (dequeue any pending requests)
			banksel		USB_buffer_data+bmRequestType
			ifclr USB_buffer_data+bmRequestType, 7, BANKED
				ifl USB_buffer_data+wLength, !=, 0
            	orif USB_buffer_data+wLengthHigh, !=, 0
					movlw		0xC8
				otherwise
					movlw		0x88
				endi
			otherwise
				movlw		0x88
			endi
			banksel		BD0OST
			movwf		BD0OST, BANKED					; set EP0 OUT UOWN back to USB and DATA0/DATA1 packet according to request type
			bcf		UCON, PKTDIS, ACCESS				; assuming there is nothing to dequeue, clear the packet disable bit
			banksel		USB_dev_req
			movlw		NO_REQUEST
			movwf		USB_dev_req, BANKED				; clear the device request in process
			movf		USB_buffer_data+bmRequestType, W, BANKED
			andlw		0x60							; extract request type bits
			select
				case STANDARD
					call		StandardRequests
					break
				case CLASS
					call		ClassRequests
					break
				case VENDOR
					call		VendorRequests
					break
				default
					bsf		USB_error_flags, 0, BANKED	; set Request Error flag
			ends
			return
;------------------------------------------------------------------------------------------------------------
StandardRequests			
			movf		USB_buffer_data+bRequest, W, BANKED
			select
				case GET_STATUS		;19/5 beoordeeld zo laten
					movf		USB_buffer_data+bmRequestType, W, BANKED
					andlw		0x1F					; extract request recipient bits
					select
						case RECIPIENT_DEVICE
							banksel		BD0IAH
							movf		BD0IAH, W, BANKED
							movwf		FSR0H, ACCESS
							movf		BD0IAL, W, BANKED			; get buffer pointer
							movwf		FSR0L, ACCESS
							banksel		USB_device_status
							movf		USB_device_status, W, BANKED	; copy device status byte to EP0 buffer
							movwf		POSTINC0
							clrf		INDF0
							banksel		BD0IBC
							movlw		0x02
							movwf		BD0IBC, BANKED			; set byte count to 2
							movlw		0xC8
							movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
							break
						case RECIPIENT_INTERFACE                                                                   
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									bsf		USB_error_flags, 0, BANKED		; set Request Error flag
									break
								case CONFIG_STATE	
									ifl USB_buffer_data+wIndex, <, NUM_INTERFACES
										banksel		BD0IAH
										movf		BD0IAH, W, BANKED
										movwf		FSR0H, ACCESS
										movf		BD0IAL, W, BANKED				; get buffer pointer
										movwf		FSR0L, ACCESS
										clrf		POSTINC0
										clrf		INDF0
										movlw		0x02
										movwf		BD0IBC, BANKED				; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED				; send packet as DATA1, set UOWN bit																								
									otherwise
										bsf		USB_error_flags, 0, BANKED		; set Request Error flag
									endi
									break
							ends
							break
						case RECIPIENT_ENDPOINT	
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
								
									
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									ifset STATUS, Z, ACCESS							; see if it is EP0
										banksel		BD0IAH
										movf		BD0IAH, W, BANKED				; put EP0 IN buffer pointer...
										movwf		FSR0H, ACCESS
										movf		BD0IAL, W, BANKED
										movwf		FSR0L, ACCESS					; ...into FSR0
										ifset UEP0, EPSTALL, ACCESS
											movlw		0x01
										otherwise
											movlw		0x00
										endi
										movwf		POSTINC0
										clrf		INDF0
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									otherwise
										bsf		USB_error_flags, 0, BANKED			; set Request Error flag
									endi
									break

								case CONFIG_STATE		;GENHID 20/5
									banksel		BD0IAH
									movf		BD0IAH, W, BANKED					; put EP0 IN buffer pointer...
									movwf		FSR0H, ACCESS
									movf		BD0IAL, W, BANKED
									movwf		FSR0L, ACCESS						; ...into FSR0
									movlw		high UEP0							; put UEP0 address...
									movwf		FSR1H, ACCESS
									movlw		low UEP0
									movwf		FSR1L, ACCESS						; ...into FSR1
									banksel		USB_buffer_data+wIndex
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP and...
									andlw		0x0F								; ...strip off direction bit
									ifclr PLUSW1, EPOUTEN, ACCESS
									andifclr PLUSW1, EPINEN, ACCESS
										bsf		USB_error_flags, 0, BANKED		; set Request Error flag
									otherwise
										ifset PLUSW1, EPSTALL, ACCESS
											movlw		0x01
										otherwise
											movlw		0x00
										endi
										movwf		POSTINC0
										clrf		INDF0
										banksel		BD0IBC
										movlw		0x02
										movwf		BD0IBC, BANKED					; set byte count to 2
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									endi
									break

								default
									bsf		USB_error_flags, 0, BANKED		; set Request Error flag
							ends
							break
						default
							bsf		USB_error_flags, 0, BANKED	; set Request Error flag
					ends
					break
				case CLEAR_FEATURE
				case SET_FEATURE
					movf		USB_buffer_data+bmRequestType, W, BANKED
					andlw		0x1F					; extract request recipient bits
					select
						case RECIPIENT_DEVICE
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case DEVICE_REMOTE_WAKEUP
									ifl USB_buffer_data+bRequest, ==, CLEAR_FEATURE
										bcf		USB_device_status, 1, BANKED
									otherwise
										bsf		USB_device_status, 1, BANKED
									endi
									banksel		BD0IBC
									clrf		BD0IBC, BANKED				; set byte count to 0
									movlw		0xC8
									movwf		BD0IST, BANKED				; send packet as DATA1, set UOWN bit
									break
								default
									bsf		USB_error_flags, 0, BANKED		; set Request Error flag
							ends
							break
						case RECIPIENT_ENDPOINT	
							movf		USB_USWSTAT, W, BANKED
							select
								case ADDRESS_STATE
									movf		USB_buffer_data+wIndex, W, BANKED		; get EP
									andlw		0x0F								; strip off direction bit
									ifset STATUS, Z, ACCESS							; see if it is EP0
										banksel		BD0IBC
										clrf		BD0IBC, BANKED					; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED					; send packet as DATA1, set UOWN bit
									otherwise
										bsf		USB_error_flags, 0, BANKED			; set Request Error flag
									endi
									break
								case CONFIG_STATE	
									movlw		high UEP0							; put UEP0 address...
									movwf		FSR0H, ACCESS
									movlw		low UEP0
									movwf		FSR0L, ACCESS						; ...into FSR0
									movlw		high BD0OST							; put BD0OST address...
									movwf		FSR1H, ACCESS
									movlw		low BD0OST
									movwf		FSR1L, ACCESS						; ...into FSR1
									movf		USB_buffer_data+wIndex, W, BANKED	; get EP
									andlw		0x0F								; strip off direction bit
									ifclr STATUS, Z, ACCESS							; if it was not EP0...
										addwf		FSR0L, F, ACCESS					; add EP number to FSR0
										ifset		STATUS, C, ACCESS
											incf		FSR0H, F, ACCESS
										endi
										rlncf		USB_buffer_data+wIndex, F, BANKED
										rlncf		USB_buffer_data+wIndex, F, BANKED
										rlncf		USB_buffer_data+wIndex, W, BANKED		; WREG now contains the proper offset into the BD table for the specified EP
										andlw		0x7C							; mask out all but the direction bit and EP number (after three left rotates)
										addwf		FSR1L, F, ACCESS					; add BD table offset to FSR1
										ifset		STATUS, C, ACCESS
											incf		FSR1H, F, ACCESS
										endi
										ifset USB_buffer_data+wIndex, 1, BANKED			; if the specified EP direction (now bit 1) is IN...
											ifset INDF0, EPINEN, ACCESS						; if the specified EP is enabled for IN transfers...
												ifl USB_buffer_data+bRequest, ==, CLEAR_FEATURE
													clrf		INDF1					; clear the stall on the specified EP
												otherwise
													movlw		0x84
													movwf		INDF1					; stall the specified EP
												endi
											otherwise
												bsf		USB_error_flags, 0, BANKED		; set Request Error flag
											endi
										otherwise										; ...otherwise the specified EP direction is OUT, so...
											ifset INDF0, EPOUTEN, ACCESS					; if the specified EP is enabled for OUT transfers...
												ifl USB_buffer_data+bRequest, ==, CLEAR_FEATURE
													movlw		0x88
													movwf		INDF1					; clear the stall on the specified EP													
												otherwise
													movlw		0x84
													movwf		INDF1					; stall the specified EP
												endi
											otherwise
												bsf			USB_error_flags, 0, BANKED		; set Request Error flag
											endi
										endi
									endi
									ifclr USB_error_flags, 0, BANKED		; if there was no Request Error...
										banksel		BD0IBC
										clrf		BD0IBC, BANKED		; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED		; send packet as DATA1, set UOWN bit
									endi
									break
								default
									bsf		USB_error_flags, 0, BANKED	; set Request Error flag
							ends
							break
						default
							bsf		USB_error_flags, 0, BANKED	; set Request Error flag
					ends
					break
				case SET_ADDRESS
					ifset USB_buffer_data+wValue, 7, BANKED		; if new device address is illegal, send Request Error
						bsf		USB_error_flags, 0, BANKED	; set Request Error flag
					otherwise
						movlw		SET_ADDRESS
						movwf		USB_dev_req, BANKED			; processing a SET_ADDRESS request
						movf		USB_buffer_data+wValue, W, BANKED
						movwf		USB_address_pending, BANKED	; save new address
						banksel		BD0IBC
						clrf		BD0IBC, BANKED				; set byte count to 0
						movlw		0xC8
						movwf		BD0IST, BANKED				; send packet as DATA1, set UOWN bit
					endi
					break
				case GET_DESCRIPTOR
					movwf		USB_dev_req, BANKED				; processing a GET_DESCRIPTOR request
					movf		USB_buffer_data+(wValue+1), W, BANKED
					select
						case DEVICE	
							movlw		low (Device-Descriptor_begin)
							movwf		USB_desc_ptr, BANKED
							call		Descriptor				; get descriptor length
							movwf		USB_bytes_left, BANKED
							ifl USB_buffer_data+(wLength+1), ==, 0
							andiff USB_buffer_data+wLength, <, USB_bytes_left
								movf		USB_buffer_data+wLength, W, BANKED
								movwf		USB_bytes_left, BANKED
							endi
							call		SendDescriptorPacket
							break
						case CONFIGURATION
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (Configuration1-Descriptor_begin)
									break
								default
									bsf		USB_error_flags, 0, BANKED	; set Request Error flag
							ends
							ifclr USB_error_flags, 0, BANKED
								addlw		0x02						; add offset for wTotalLength
								movwf		USB_desc_ptr, BANKED
								call		Descriptor					; get total descriptor length
								movwf		USB_bytes_left, BANKED
								movlw		0x02
								subwf		USB_desc_ptr, F, BANKED			; subtract offset for wTotalLength
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							endi
							break
						case STRING		;TOEVOEGEN STRING255 gedaan, default error verwijderd
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (String0-Descriptor_begin)
									break
								case 1
									movlw		low (String1-Descriptor_begin)
									break
								case 2
									movlw		low (String2-Descriptor_begin)
									break
								case 3
				                        	movlw   	low (String3-Descriptor_begin)
				                        	break
				                      	default
				                        	movlw   	low (String255-Descriptor_begin)
				                        	break
	
							ends
							ifclr USB_error_flags, 0, BANKED
								movwf		USB_desc_ptr, BANKED
								call		Descriptor					; get descriptor length
								movwf		USB_bytes_left, BANKED
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							endi
							break
						case HID
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		low (HID1-Descriptor_begin)
									break
								default
									bsf		USB_error_flags, 0, BANKED	; set Request Error flag
							ends
							ifclr USB_error_flags, 0, BANKED
								movwf		USB_desc_ptr, BANKED
								call		Descriptor				; get descriptor length
								movwf		USB_bytes_left, BANKED
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							endi
							break
						case REPORT		;2F is reportlengte
							movf		USB_buffer_data+wValue, W, BANKED
							select
								case 0
									movlw		Endreport1-Report1
									movwf		USB_bytes_left, BANKED	; set report descriptor length
									movlw		low (Report1-Descriptor_begin)
									break
								default
									bsf		USB_error_flags, 0, BANKED	; set Request Error flag
							ends
							ifclr USB_error_flags, 0, BANKED
								movwf		USB_desc_ptr, BANKED
								ifl USB_buffer_data+(wLength+1), ==, 0
								andiff USB_buffer_data+wLength, <, USB_bytes_left
									movf		USB_buffer_data+wLength, W, BANKED
									movwf		USB_bytes_left, BANKED
								endi
								call		SendDescriptorPacket
							endi
							break
						default
							bsf		USB_error_flags, 0, BANKED	; set Request Error flag
					ends
					break
				case GET_CONFIGURATION
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS
					banksel		USB_curr_config
					movf		USB_curr_config, W, BANKED
					movwf		INDF0					; copy current device configuration to EP0 IN buffer
					banksel		BD0IBC
					movlw		0x01
					movwf		BD0IBC, BANKED			; set EP0 IN byte count to 1
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_CONFIGURATION
					ifl USB_buffer_data+wValue, <=, NUM_CONFIGURATIONS
						clrf		UEP1, ACCESS		; clear all EP control registers except for EP0 to disable EP1-EP15 prior to setting configuration
						clrf		UEP2, ACCESS
						clrf		UEP3, ACCESS
						clrf		UEP4, ACCESS
						clrf		UEP5, ACCESS
						clrf		UEP6, ACCESS
						clrf		UEP7, ACCESS
						clrf		UEP8, ACCESS
						clrf		UEP9, ACCESS
						clrf		UEP10, ACCESS
						clrf		UEP11, ACCESS
						clrf		UEP12, ACCESS
						clrf		UEP13, ACCESS
						clrf		UEP14, ACCESS
						clrf		UEP15, ACCESS
						movf		USB_buffer_data+wValue, W, BANKED
						movwf		USB_curr_config, BANKED
						select
							case 0
								movlw		ADDRESS_STATE
								movwf		USB_USWSTAT, BANKED
								break
							default
								movlw		CONFIG_STATE
								movwf		USB_USWSTAT, BANKED
								banksel		BD1OBC
								movlw		MAX_packet_size
								movwf		BD1OBC, BANKED			; set EP1 OUT byte count to MAX_packet_size
								movlw		low (USB_Buffer+2*EP0_packet_size)	; EP1 OUT gets a buffer...
								movwf		BD1OAL, BANKED
								movlw		high (USB_Buffer+2*EP0_packet_size)
								movwf		BD1OAH, BANKED			; ...set its address
								movlw		0x88
								movwf		BD1OST, BANKED			; set UOWN bit (USB can write EP1 OUT buffer)
								movlw		MAX_packet_size
								movwf		BD1IBC, BANKED			; set EP1 IN byte count to MAX_packet_size 
								movlw		low (USB_Buffer+2*EP0_packet_size+MAX_packet_size)	; EP1 IN gets a buffer...
								movwf		BD1IAL, BANKED
								movlw		high (USB_Buffer+2*EP0_packet_size+MAX_packet_size)
								movwf		BD1IAH, BANKED			; ...set its address
								movlw		0x48
								movwf		BD1IST, BANKED			; clear UOWN bit (PIC can write EP1 IN buffer)
								movlw		ENDPT_NON_CONTROL
								movwf		UEP1, ACCESS			; enable EP1 for interrupt in transfers
						ends
						banksel		BD0IBC
						clrf		BD0IBC, BANKED			; set byte count to 0
						movlw		0xC8
						movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					otherwise
						bsf		USB_error_flags, 0, BANKED	; set Request Error flag
					endi
					break
				case GET_INTERFACE
					movf		USB_USWSTAT, W, BANKED
					select
						case CONFIG_STATE
							ifl USB_buffer_data+wIndex, <, NUM_INTERFACES
								banksel		BD0IAH
								movf		BD0IAH, W, BANKED
								movwf		FSR0H, ACCESS
								movf		BD0IAL, W, BANKED		; get buffer pointer
								movwf		FSR0L, ACCESS
								clrf		INDF0					; always send back 0 for bAlternateSetting
								movlw		0x01
								movwf		BD0IBC, BANKED			; set byte count to 1
								movlw		0xC8
								movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
							otherwise
								bsf		USB_error_flags, 0, BANKED	; set Request Error flag
							endi
							break
						default
							bsf		USB_error_flags, 0, BANKED	; set Request Error flag
					ends
					break
				case SET_INTERFACE
					movf		USB_USWSTAT, W, BANKED
					select
						case CONFIG_STATE
							ifl USB_buffer_data+wIndex, <, NUM_INTERFACES
								movf		USB_buffer_data+wValue, W, BANKED
								select
									case 0									; currently support only bAlternateSetting of 0
										banksel		BD0IBC
										clrf		BD0IBC, BANKED			; set byte count to 0
										movlw		0xC8
										movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
										break
									default
										bsf		USB_error_flags, 0, BANKED	; set Request Error flag
								ends
							otherwise
								bsf		USB_error_flags, 0, BANKED	; set Request Error flag
							endi
							break
						default
							bsf		USB_error_flags, 0, BANKED	; set Request Error flag
					ends
					break
				case SET_DESCRIPTOR
				case SYNCH_FRAME
				default
					bsf		USB_error_flags, 0, BANKED	; set Request Error flag
					break
			ends
			return
;----------------------------------------------------------------------------------------------------------------------
ClassRequests
			movf		USB_buffer_data+bRequest, W, BANKED
			select
				case GET_REPORT
					movf		USB_buffer_data+wValueHigh, W, BANKED
					select
						case INPUT
							banksel		BD0IAH
							movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
							movwf		FSR0H, ACCESS
							movf		BD0IAL, W, BANKED
							movwf		FSR0L, ACCESS			; ...into FSR0
							banksel		Ctrl_io_buffer
							movf		Ctrl_io_buffer, W, BANKED	; copy first byte...
							movwf		POSTINC0				; ...to EP0 IN buffer
							movf		Ctrl_io_buffer+1, W, BANKED	; copy second byte...
							movwf		INDF0					; ...to EP0 IN buffer
							banksel		BD0IBC
							movlw		0x02
							movwf		BD0IBC, BANKED			; set EP0 IN buffer byte count to 2
							movlw		0xC8
							movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
							break
						case FEATURE
							banksel		BD0IAH
							movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
							movwf		FSR0H, ACCESS
							movf		BD0IAL, W, BANKED
							movwf		FSR0L, ACCESS			; ...into FSR0
							banksel		Feature_buffer
							movf		Feature_buffer, W, BANKED	; copy first byte...
							movwf		POSTINC0				; ...to EP0 IN buffer
							movf		Feature_buffer+1, W, BANKED	; copy second byte...
							movwf		INDF0					; ...to EP0 IN buffer
							banksel		BD0IBC
							movlw		0x02
							movwf		BD0IBC, BANKED			; set EP0 IN buffer byte count to 2
							movlw		0xC8
							movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
							break
						default
							bsf		USB_error_flags, 0, BANKED	; set Request Error flag
					ends
			
					
					break

				case SET_REPORT
					movf		USB_buffer_data+wValue, W, BANKED
					movwf		USB_protocol, BANKED		; update the new protocol value
					banksel		BD0IBC
					clrf		BD0IBC, BANKED			; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break

				case GET_PROTOCOL
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS			; ...into FSR0
					banksel		USB_protocol
					movf		USB_protocol, W, BANKED
					movwf		INDF0
					banksel		BD0IBC
					movlw		0x01
					movwf		BD0IBC, BANKED			; set byte count to 1
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_PROTOCOL
					movf		USB_buffer_data+wValue, W, BANKED
					movwf		USB_protocol, BANKED	; update the new protocol value
					banksel		BD0IBC
					clrf		BD0IBC, BANKED			; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case GET_IDLE
					banksel		BD0IAH
					movf		BD0IAH, W, BANKED			; put EP0 IN buffer pointer...
					movwf		FSR0H, ACCESS
					movf		BD0IAL, W, BANKED
					movwf		FSR0L, ACCESS			; ...into FSR0
					banksel		USB_idle_rate
					movf		USB_idle_rate, W, BANKED
					movwf		INDF0
					banksel		BD0IBC
					movlw		0x01
					movwf		BD0IBC, BANKED			; set byte count to 1
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				case SET_IDLE
					movf		USB_buffer_data+wValueHigh, W, BANKED
					movwf		USB_idle_rate, BANKED	; update the new idle rate
					banksel		BD0IBC
					clrf		BD0IBC, BANKED			; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED			; send packet as DATA1, set UOWN bit
					break
				default
					bsf		USB_error_flags, 0, BANKED	; set Request Error flag
			ends
			return
;------------------------------------------------------------------------------------------------------------
VendorRequests
			bsf		USB_error_flags, 0, BANKED	; set Request Error flag
			return
;------------------------------------------------------------------------------------------------------------------
ProcessInToken
			banksel		USB_USTAT
			movf		USB_USTAT, W, BANKED
			andlw		0x18		; extract the EP bits
			select
				case EP0
					movf		USB_dev_req, W, BANKED
					select
						case SET_ADDRESS
							movf		USB_address_pending, W, BANKED
							movwf		UADDR, ACCESS
							select
								case 0
									movlw		DEFAULT_STATE
									movwf		USB_USWSTAT, BANKED
									break
								default
									movlw		ADDRESS_STATE
									movwf		USB_USWSTAT, BANKED
							ends
							break
						case GET_DESCRIPTOR
							call		SendDescriptorPacket
							break
					ends
					break
				case EP1
					break
				case EP2
					break
			ends
			return
;-----------------------------------------------------------------------------------------------------------
ProcessOutToken
			banksel		USB_USTAT
			movf		USB_USTAT, W, BANKED
			andlw		0x18		; extract the EP bits
			select
				case EP0
							
					movf		USB_dev_req, W, BANKED
					select
						case SET_REPORT
							movlw		NO_REQUEST
							movwf		USB_dev_req, BANKED	; clear device request
							banksel		BD0OBC
							ifl BD0OBC, ==, 0x02
								movf		BD0OAH, W, BANKED	; put EP0 OUT buffer pointer...
								movwf		FSR0H, ACCESS
								movf		BD0OAL, W, BANKED
								movwf		FSR0L, ACCESS		; ...into FSR0
								banksel		Report_type
								movf		Report_type, W, BANKED
								select
									case OUTPUT
										movf		POSTINC0, W			; get the first byte in the buffer and...
										movwf		Ctrl_io_buffer, BANKED	; ...put it in the conrtol I/O report buffer
										movf		INDF0, W			; get the second byte in the buffer and...
										movwf		Ctrl_io_buffer+1, BANKED	; ...put it the control I/O report buffer
										break
									case FEATURE
										movf		POSTINC0, W			; get the first byte in the buffer and...
										movwf		Feature_buffer, BANKED	; ...put it in the feature report buffer
										movf		INDF0, W			; get the second byte in the buffer and...
										movwf		Feature_buffer+1, BANKED	; ...put it the feature report buffer
										break
									default
										bsf		USB_error_flags, 0, BANKED	; set Request Error flag
								ends
							otherwise
								bsf		USB_error_flags, 0, BANKED	; set Request Error flag
							endi
					ends
					banksel		BD0OBC
					movlw		EP0_packet_size
					movwf		BD0OBC, BANKED
					movlw		0x88
					movwf		BD0OST, BANKED
					clrf		BD0IBC, BANKED		; set byte count to 0
					movlw		0xC8
					movwf		BD0IST, BANKED		; send packet as DATA1, set UOWN bit
					break

				case EP1
					break
				case EP2
					break
			ends
			return
;---------------------------------------------------------------------------------------------------------------------
Descriptor
			movlw		upper Descriptor_begin
			movwf		TBLPTRU, ACCESS
			movlw		high Descriptor_begin
			movwf		TBLPTRH, ACCESS
			movlw		low Descriptor_begin
			banksel		USB_desc_ptr
			addwf		USB_desc_ptr, W, BANKED
			bnc		$+8
			incf		TBLPTRH, F
			bnz		$+4
			incf		TBLPTRU, F
			movwf		TBLPTRL, ACCESS
			tblrd*
			movf		TABLAT, W
			return
;----------------------------------------------------------------------------------------------------------			
SendDescriptorPacket
			banksel		USB_bytes_left
			ifl USB_bytes_left, <, EP0_packet_size
				movlw		NO_REQUEST
				movwf		USB_dev_req, BANKED		; sending a short packet, so clear device request
				movf		USB_bytes_left, W, BANKED
			otherwise
				movlw		EP0_packet_size
			endi
			subwf		USB_bytes_left, F, BANKED
			movwf		USB_packet_length, BANKED
			banksel		BD0IBC
			movwf		BD0IBC, BANKED			; set EP0 IN byte count with packet size
			movf		BD0IAH, W, BANKED		; put EP0 IN buffer pointer...
			movwf		FSR0H, ACCESS
			movf		BD0IAL, W, BANKED
			movwf		FSR0L, ACCESS			; ...into FSR0
			banksel		USB_loop_index
			forlf USB_loop_index, 1, USB_packet_length
				call		Descriptor			; get next byte of descriptor being sent
				movwf		POSTINC0			; copy to EP0 IN buffer, and increment FSR0
				incf		USB_desc_ptr, F, BANKED	; increment the descriptor pointer
			next USB_loop_index
			banksel		BD0IST
			movlw		0x40
			xorwf		BD0IST, W, BANKED		; toggle the DATA01 bit
			andlw		0x40					; clear the PIDs bits
			iorlw		0x88					; set UOWN and DTS bits
			movwf		BD0IST, BANKED
			return

			end
